from typing import Any

import pytest
from deepdiff import DeepDiff

import pipelime.choixe.ast.nodes as ast
from pipelime.choixe.visitors import unparse


@pytest.mark.parametrize(
    ["node", "expected"],
    [
        [ast.LiteralNode(data=10), 10],
        [ast.LiteralNode(data=0.124), 0.124],
        [ast.LiteralNode(data="hello"), "hello"],
        [ast.LiteralNode(data="my_var"), "my_var"],
        [
            ast.VarNode(identifier=ast.LiteralNode(data="variable.one")),
            "$var(variable.one)",
        ],
        [
            ast.VarNode(
                identifier=ast.LiteralNode(data="variable.one"),
                env=ast.LiteralNode(data=True),
            ),
            "$var(variable.one, env=True)",
        ],
        [
            ast.VarNode(
                identifier=ast.LiteralNode(data="variable.one"),
                default=ast.LiteralNode(data=-24),
            ),
            "$var(variable.one, default=-24)",
        ],
        [
            ast.VarNode(
                identifier=ast.LiteralNode(data="variable.one"),
                default=ast.LiteralNode(data=-24),
                env=ast.LiteralNode(data=True),
            ),
            "$var(variable.one, default=-24, env=True)",
        ],
        [
            ast.VarNode(
                identifier=ast.LiteralNode(data="variable"),
                default=ast.LiteralNode(data=32),
                help=ast.LiteralNode(data="help message"),
            ),
            '$var(variable, default=32, help="help message")',
        ],
        [
            ast.ImportNode(path=ast.LiteralNode(data="path/to/file.yaml")),
            '$import("path/to/file.yaml")',
        ],
        [
            ast.SweepNode(
                ast.LiteralNode(data="a"),
                ast.LiteralNode(data="variable"),
                ast.LiteralNode(data=10),
            ),
            "$sweep(a, variable, 10)",
        ],
        [ast.StrBundleNode(ast.LiteralNode(data="alice")), "alice"],
        [
            ast.StrBundleNode(
                ast.LiteralNode(data="alice "),
                ast.VarNode(
                    identifier=ast.LiteralNode(data="foo"),
                    default=ast.LiteralNode(data="loves"),
                ),
                ast.LiteralNode(data=" bob"),
            ),
            "alice $var(foo, default=loves) bob",
        ],
        [
            ast.DictNode(
                nodes={
                    ast.LiteralNode(data="key1"): ast.LiteralNode(data=10),
                    ast.LiteralNode(data="key2"): ast.DictNode(
                        nodes={
                            ast.LiteralNode(data="key1"): ast.LiteralNode(data=10.2),
                            ast.LiteralNode(data="key2"): ast.LiteralNode(data="hello"),
                        }
                    ),
                }
            ),
            {"key1": 10, "key2": {"key1": 10.2, "key2": "hello"}},
        ],
        [
            ast.DictNode(
                nodes={
                    ast.StrBundleNode(
                        ast.VarNode(identifier=ast.LiteralNode(data="var")),
                        ast.LiteralNode(data="foo"),
                    ): ast.LiteralNode(data="bar")
                }
            ),
            {"$var(var)foo": "bar"},
        ],
        [
            ast.ListNode(
                ast.LiteralNode(data=10),
                ast.LiteralNode(data=-0.25),
                ast.ListNode(ast.LiteralNode(data="aa")),
            ),
            [10, -0.25, ["aa"]],
        ],
        [
            ast.SymbolNode(symbol=ast.LiteralNode(data="numpy.zeros")),
            "$symbol(numpy.zeros)",
        ],
        [
            ast.InstanceNode(
                symbol=ast.LiteralNode(data="path/to_my/file.py:MyClass"),
                args=ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="arg1"): ast.InstanceNode(
                            symbol=ast.LiteralNode(data="module.submodule.function"),
                            args=ast.DictNode(
                                nodes={
                                    ast.LiteralNode(data="a"): ast.ListNode(
                                        ast.LiteralNode(data=1), ast.LiteralNode(data=2)
                                    ),
                                    ast.LiteralNode(data="b"): ast.LiteralNode(
                                        data=100
                                    ),
                                }
                            ),
                        )
                    }
                ),
            ),
            {
                "$call": "path/to_my/file.py:MyClass",
                "$args": {
                    "arg1": {
                        "$call": "module.submodule.function",
                        "$args": {"a": [1, 2], "b": 100},
                    }
                },
            },
        ],
        [
            ast.ModelNode(
                symbol=ast.LiteralNode(data="path/to_my/file.py:MyModel"),
                args=ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="arg1"): ast.DictNode(
                            nodes={
                                ast.LiteralNode(data="a"): ast.ListNode(
                                    ast.LiteralNode(data=1), ast.LiteralNode(data=2)
                                ),
                                ast.LiteralNode(data="b"): ast.LiteralNode(data=100),
                            }
                        ),
                    }
                ),
            ),
            {
                "$model": "path/to_my/file.py:MyModel",
                "$args": {"arg1": {"a": [1, 2], "b": 100}},
            },
        ],
        [
            ast.ForNode(
                iterable=ast.LiteralNode(data="my.var"),
                body=ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="Hello"): ast.LiteralNode(data="World"),
                        ast.StrBundleNode(
                            ast.LiteralNode(data="Number_"),
                            ast.IndexNode(identifier=ast.LiteralNode(data="x")),
                        ): ast.ItemNode(identifier=ast.LiteralNode(data="x")),
                    }
                ),
                identifier=ast.LiteralNode(data="x"),
            ),
            {"$for(my.var, x)": {"Hello": "World", "Number_$index(x)": "$item(x)"}},
        ],
        [
            ast.ForNode(
                iterable=ast.LiteralNode(data="my.var"),
                body=ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="Hello"): ast.LiteralNode(data="World"),
                        ast.StrBundleNode(
                            ast.LiteralNode(data="Number_"), ast.IndexNode()
                        ): ast.ItemNode(),
                    }
                ),
            ),
            {"$for(my.var)": {"Hello": "World", "Number_$index": "$item"}},
        ],
        [
            ast.SwitchNode(
                value=ast.LiteralNode(data="nation"),
                cases=[
                    (
                        ast.ListNode(
                            ast.LiteralNode(data="UK"),
                            ast.LiteralNode(data="USA"),
                            ast.LiteralNode(data="Australia"),
                        ),
                        ast.LiteralNode(data="hello"),
                    ),
                    (
                        ast.LiteralNode(data="Italy"),
                        ast.LiteralNode(data="ciao"),
                    ),
                ],
                default=ast.LiteralNode(data="*raise your hand*"),
            ),
            {
                "$switch(nation)": [
                    {
                        "$case": ["UK", "USA", "Australia"],
                        "$then": "hello",
                    },
                    {
                        "$case": "Italy",
                        "$then": "ciao",
                    },
                    {
                        "$default": "*raise your hand*",
                    },
                ]
            },
        ],
        [ast.UuidNode(), "$uuid"],
        [ast.DateNode(), "$date"],
        [ast.DateNode(format=ast.LiteralNode(data="%Y%m%d")), '$date("%Y%m%d")'],
        [ast.CmdNode(command=ast.LiteralNode(data="ls -lha")), '$cmd("ls -lha")'],
        [ast.TmpDirNode(), "$tmp"],
        [ast.TmpDirNode(name=ast.LiteralNode(data="my_tmp")), "$tmp(my_tmp)"],
        [
            ast.DictBundleNode(
                ast.ForNode(
                    iterable=ast.LiteralNode(data="alpha"),
                    body=ast.DictNode(
                        nodes={
                            ast.StrBundleNode(
                                ast.LiteralNode(data="node_"), ast.IndexNode()
                            ): ast.StrBundleNode(
                                ast.LiteralNode(data="Hello_"), ast.ItemNode()
                            )
                        }
                    ),
                ),
                ast.ForNode(
                    iterable=ast.LiteralNode(data="beta"),
                    body=ast.DictNode(
                        nodes={
                            ast.StrBundleNode(
                                ast.LiteralNode(data="node_"), ast.IndexNode()
                            ): ast.StrBundleNode(
                                ast.LiteralNode(data="Ciao_"), ast.ItemNode()
                            )
                        }
                    ),
                ),
                ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="a"): ast.LiteralNode(data=10),
                        ast.LiteralNode(data="b"): ast.DictNode(
                            nodes={
                                ast.LiteralNode(data="c"): ast.LiteralNode(data=10.0),
                                ast.LiteralNode(data="d"): ast.LiteralNode(
                                    data="hello"
                                ),
                            }
                        ),
                    }
                ),
            ),
            {
                "$for(alpha)": {"node_$index": "Hello_$item"},
                "$for(beta)": {"node_$index": "Ciao_$item"},
                "a": 10,
                "b": {"c": 10.0, "d": "hello"},
            },
        ],
    ],
)
def test_unparse(node: ast.Node, expected: Any):
    assert not DeepDiff(unparse(node), expected)



from pipelime.sequences import SamplesSequence
import typing as t
from torch.utils.data import Dataset  # type: ignore


class TorchDataset(Dataset[t.Mapping[str, t.Any]]):
    def __init__(self, sequence: SamplesSequence):
        self._sequence = sequence

    def __len__(self) -> int:
        return len(self._sequence)

    def __getitem__(self, idx: int) -> t.Mapping[str, t.Any]:
        return self._sequence[idx].to_dict()



import typing as t

import pipelime.sequences as pls


class TestSamplesSequenceOperations:
    def test_map(self, minimnist_dataset: dict):
        import pipelime.items as pli
        import pipelime.stages as plst

        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        ).map(
            plst.StageLambda(
                lambda x: pls.Sample(
                    {k: pli.JsonMetadataItem({"the_answer": 42}) for k in x}
                )
            )
        )
        for sample in source:
            assert all(k in sample for k in minimnist_dataset["item_keys"])
            for k, v in sample.items():
                assert k in minimnist_dataset["item_keys"]
                assert isinstance(v, pli.JsonMetadataItem)

                raw = v()
                assert isinstance(raw, t.Mapping)
                assert "the_answer" in raw
                assert raw["the_answer"] == 42

    def test_zip(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )

        source_nkeys = len(minimnist_dataset["item_keys"])
        keys1 = minimnist_dataset["item_keys"][: source_nkeys // 2]
        keys2 = minimnist_dataset["item_keys"][source_nkeys // 2 :]  # noqa: E203
        seq1 = pls.SamplesSequence.from_list(  # type: ignore
            [sample.extract_keys(*keys1) for sample in source]
        )
        seq2 = pls.SamplesSequence.from_list(  # type: ignore
            [sample.extract_keys(*keys2) for sample in source]
        )

        prefix = "prefix*"
        zipped = seq1.zip(seq2, key_format=prefix)
        assert len(zipped) == len(seq1)
        assert len(zipped) == len(seq2)
        for sample_m, sample_1, sample_2, sample_s in zip(zipped, seq1, seq2, source):
            assert len(sample_1.keys() & sample_2.keys()) == 0
            assert len(sample_1.keys() | sample_2.keys()) == len(sample_s.keys())
            assert len(sample_s.keys()) == len(sample_m.keys())
            assert all(k in sample_m for k in sample_1)
            assert all(prefix.replace("*", k) in sample_m for k in sample_2)
            assert all(
                (prefix.replace("*", k) if k in sample_2 else k) in sample_m
                for k in sample_s
            )
            assert all(
                (k[len(prefix) - 1 :] if k.startswith(prefix[:-1]) else k)  # noqa: E203
                in sample_s
                for k in sample_m
            )
            assert all(v is sample_m[k] for k, v in sample_1.items())
            assert all(
                v is sample_m[prefix.replace("*", k)] for k, v in sample_2.items()
            )
            assert all(
                v is sample_m[(prefix.replace("*", k) if k in sample_2 else k)]
                for k, v in sample_s.items()
            )
            assert all(
                v
                is sample_s[
                    (
                        k[len(prefix) - 1 :]  # noqa: E203
                        if k.startswith(prefix[:-1])
                        else k
                    )
                ]
                for k, v in sample_m.items()
            )

    def _cat_test(self, minimnist_dataset: dict, fn):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        double_source = fn(source)

        length = len(source)
        assert 2 * length == len(double_source)
        for src_idx in range(length):
            assert source[src_idx] is double_source[src_idx]
            assert source[src_idx] is double_source[src_idx + length]

    def test_cat(self, minimnist_dataset: dict):
        self._cat_test(minimnist_dataset, lambda x: x.cat(x))

    def test_add(self, minimnist_dataset: dict):
        self._cat_test(minimnist_dataset, lambda x: x + x)

    def test_filter(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        filtered = source.filter(lambda x: x["label"] == 0)

        raw_count = len([sample for sample in source if sample["label"] == 0])
        assert raw_count == len(filtered)

        for sample in filtered:
            assert sample["label"] == 0

    def test_sort(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        ).sort(lambda x: -1 * x["label"]())

        last_label = 9
        for sample in source:
            assert sample["label"]() <= last_label
            last_label = sample["label"]()

    def test_slice(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        start, stop, step = 4, 13, 3
        sliced = source.slice(start=start, stop=stop, step=step)
        raw_sliced = [source[idx] for idx in range(start, stop, step)]

        assert len(sliced) == len(raw_sliced)
        for s1, s2 in zip(sliced, raw_sliced):
            assert s1 is s2

    def test_slice_get(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        start, stop, step = 7, 18, 2
        sliced = source[start:stop:step]
        raw_sliced = [source[idx] for idx in range(start, stop, step)]

        assert len(sliced) == len(raw_sliced)
        for s1, s2 in zip(sliced, raw_sliced):
            assert s1 is s2

    def test_shuffle(self, minimnist_dataset: dict):
        seed = 42
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        shuffled_1 = source.shuffle(seed=seed)
        shuffled_2 = source.shuffle(seed=seed)
        shuffled_3 = source.shuffle(seed=seed + 1)

        length = len(source)
        assert length == len(shuffled_1)
        assert length == len(shuffled_2)
        assert length == len(shuffled_3)
        assert any(source[idx] is not shuffled_1[idx] for idx in range(length))
        assert any(source[idx] is not shuffled_2[idx] for idx in range(length))
        assert any(source[idx] is not shuffled_3[idx] for idx in range(length))
        assert all(shuffled_1[idx] is shuffled_2[idx] for idx in range(length))
        assert any(shuffled_1[idx] is not shuffled_3[idx] for idx in range(length))

    def test_enumerate(self, minimnist_dataset: dict):
        import pipelime.items as pli

        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        enum_seq = source.enumerate(
            idx_key="custom_id", item_cls="pipelime.items.TxtNumpyItem"
        )

        for (idx, s_sample), e_sample in zip(enumerate(source), enum_seq):
            assert "custom_id" in e_sample
            assert isinstance(e_sample["custom_id"], pli.TxtNumpyItem)
            assert int(e_sample["custom_id"]()) == idx  # type: ignore
            assert all(v == e_sample[k] for k, v in s_sample.items())

    def test_repeat(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        repeat_seq = source.repeat(3)

        assert len(repeat_seq) == 3 * len(source)

        riter = iter(repeat_seq)
        for i in range(3):
            for s in source:
                assert next(riter) is s

    def test_select(self, minimnist_dataset: dict):
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=minimnist_dataset["path"], merge_root_items=False
        )
        idxs = [1, 2, 5, 9]
        select_seq = source.select(idxs)

        assert len(select_seq) == len(idxs)

        for i, s in zip(idxs, select_seq):
            assert s is source[i]

    def test_cache(self, minimnist_dataset: dict):
        import numpy as np

        from pipelime.stages import SampleStage

        class LocalStage(SampleStage):
            counter: int = 0

            def __call__(self, x):
                print("counter", self.counter)
                self.counter += 1
                x = x.duplicate_key("image", "img2")
                _ = x["img2"]()
                return x

        local_stage = LocalStage()
        source = (
            pls.SamplesSequence.from_underfolder(  # type: ignore
                folder=minimnist_dataset["path"], merge_root_items=False
            )
            .map(local_stage)
            .cache()
        )

        saved_samples = [x for x in source]
        for idx, x in enumerate(source):
            s = saved_samples[idx]
            assert x is not s
            assert x["image"]().size != 0
            assert x["img2"]().size != 0
            assert np.array_equal(x["image"](), s["image"](), equal_nan=True)
            assert np.array_equal(x["img2"](), s["img2"](), equal_nan=True)
        assert local_stage.counter == len(source)



import pytest
from pathlib import Path
import pipelime.sequences as pls

import typing as t


class TestSamplesSequenceWriters:
    def _read_write_data(
        self, source_dataset, out_folder, **writer_kwargs
    ) -> t.Tuple[pls.SamplesSequence, pls.SamplesSequence]:
        source = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=source_dataset["path"], merge_root_items=True
        )
        for _ in source.to_underfolder(folder=out_folder, **writer_kwargs):
            pass
        dest = pls.SamplesSequence.from_underfolder(  # type: ignore
            folder=out_folder, merge_root_items=True
        )
        return source, dest

    def _check_data(self, source: pls.SamplesSequence, dest: pls.SamplesSequence):
        import pipelime.items as pli
        import numpy as np

        assert len(source) == len(dest)
        for s1, s2 in zip(source, dest):
            assert s1.keys() == s2.keys()
            for k, v1 in s1.items():
                v2 = s2[k]
                assert v1.__class__ == v2.__class__
                if isinstance(v1, pli.NumpyItem):
                    assert np.array_equal(v1(), v2(), equal_nan=True)  # type: ignore
                else:
                    assert v1() == v2()

    def test_to_underfolder(self, minimnist_dataset: dict, tmp_path: Path):
        source, dest = self._read_write_data(minimnist_dataset, tmp_path / "outfolder")
        self._check_data(source, dest)

    @pytest.mark.parametrize("exists_ok", [True, False])
    def test_to_underfolder_exists(self, tmp_path: Path, exists_ok: bool):
        tmp_path = tmp_path / "outfolder" / "data"
        tmp_path.mkdir(parents=True)

        try:
            _ = pls.SamplesSequence.from_list([]).to_underfolder(  # type: ignore
                folder=tmp_path, exists_ok=exists_ok
            )
        except FileExistsError:
            assert not exists_ok
            return
        assert exists_ok

    def test_to_underfolder_serialization_mode(
        self, minimnist_private_dataset: dict, tmp_path: Path
    ):
        import pipelime.items as pli

        item_key = minimnist_private_dataset["item_keys"][0]
        with pli.item_serialization_mode(pli.SerializationMode.HARD_LINK):
            source, dest = self._read_write_data(
                minimnist_private_dataset,
                tmp_path / "outfolder",
                key_serialization_mode={item_key: pli.SerializationMode.DEEP_COPY},
            )
        self._check_data(source, dest)

        for sample in dest:
            for key, item in sample.items():
                assert isinstance(item._file_sources, t.Sequence)
                assert len(item._file_sources) == 1
                path = Path(item._file_sources[0])
                assert not path.is_symlink()
                assert path.is_file()
                assert path.stat().st_nlink == (1 if key == item_key else 2)

    def test_deep_copy(self, minimnist_dataset: dict, tmp_path: Path):
        import pipelime.items as pli

        with pli.item_serialization_mode(pli.SerializationMode.DEEP_COPY):
            source, dest = self._read_write_data(
                minimnist_dataset, tmp_path / "outfolder"
            )
        self._check_data(source, dest)

        for sample in dest:
            for item in sample.values():
                assert isinstance(item._file_sources, t.Sequence)
                assert len(item._file_sources) == 1
                path = Path(item._file_sources[0])
                assert not path.is_symlink()
                assert path.is_file()
                assert path.stat().st_nlink == 1

    def test_symlink(self, minimnist_dataset: dict, tmp_path: Path):
        import pipelime.items as pli
        import platform

        with pli.item_serialization_mode(pli.SerializationMode.SYM_LINK):
            source, dest = self._read_write_data(
                minimnist_dataset, tmp_path / "outfolder"
            )
        self._check_data(source, dest)

        for sample in dest:
            for item in sample.values():
                assert isinstance(item._file_sources, t.Sequence)
                assert len(item._file_sources) == 1
                path = Path(item._file_sources[0])
                assert path.is_file()
                assert not path.is_symlink()

        if not platform.system() == "Windows":
            for path in (tmp_path / "outfolder").rglob("*"):
                assert not path.is_file() or path.is_symlink()

    def test_hardlink(self, minimnist_private_dataset: dict, tmp_path: Path):
        import pipelime.items as pli

        with pli.item_serialization_mode(pli.SerializationMode.HARD_LINK):
            source, dest = self._read_write_data(
                minimnist_private_dataset, tmp_path / "outfolder"
            )
        self._check_data(source, dest)

        for sample in dest:
            for item in sample.values():
                assert isinstance(item._file_sources, t.Sequence)
                assert len(item._file_sources) == 1
                path = Path(item._file_sources[0])
                assert not path.is_symlink()
                assert path.is_file()
                assert path.stat().st_nlink == 2

    def test_create_new_file(self, minimnist_dataset: dict, tmp_path: Path):
        import pipelime.items as pli

        with pli.item_serialization_mode(
            pli.SerializationMode.CREATE_NEW_FILE, pli.MetadataItem, pli.NumpyItem
        ):
            source, dest = self._read_write_data(
                minimnist_dataset, tmp_path / "outfolder"
            )
        self._check_data(source, dest)

        for sample in dest:
            for item in sample.values():
                assert isinstance(item._file_sources, t.Sequence)
                assert len(item._file_sources) == 1
                path = Path(item._file_sources[0])
                assert not path.is_symlink()
                assert path.is_file()
                assert path.stat().st_nlink == 1



import typing as t


class _RefItem:
    def __init__(self, parent, idx):
        self._parent = parent
        self._key = idx

    def __repr__(self):
        return f"_RefItem({self._parent}, {self._key})"

    def __getitem__(self, key):
        return self._parent[key] if self._key is None else self._parent[self._key][key]

    def __setitem__(self, key, val):
        if self._key is None:
            self._parent[key] = val
        else:
            self._parent[self._key][key] = val

    def __contains__(self, key):
        return (
            key in self._parent if self._key is None else key in self._parent[self._key]
        )

    def __len__(self):
        return len(self._parent) if self._key is None else len(self._parent[self._key])

    def extend(self, val):
        self._parent.extend(val) if self._key is None else self._parent[
            self._key
        ].extend(val)

    def is_none(self):
        return (
            (self._parent is None)
            if self._key is None
            else (self._parent[self._key] is None)
        )

    def set_(self, val):
        if self._key is None:
            self._parent = val
        else:
            self._parent[self._key] = val

    def get_(self):
        return self._parent if self._key is None else self._parent[self._key]


def deep_set_(  # noqa: C901
    obj: t.Union[t.Sequence, t.Mapping],
    key_path: t.Union[str, t.Sequence],
    value: t.Any,
    append: bool = False,
    default_sequence_factory=list,
    default_mapping_factory=dict,
):
    """Sets value based on full path key, ie, 'a.b.0.d' or 'a.b[0].d'. Note that the `.`
    refers to mapping keys of type string, while the `[]` indexes list elements.

    Args:
        obj (Union[Sequence, Mapping]): The object to update in place.
        key_path (Union[str, Sequence]): Full path key in pydash notation.
        value (Any): The value to set.
        append (bool, optional): If the key path is present, append the value instead of
            overwriting the current one. Defaults to False.
        default_sequence_factory (optional): a callable to create a new sequence object.
            Defaults to list.
        default_mapping_factory (optional): a callable to create a new mapping object.
            Defaults to dict.
        Defaults to True.
    """
    from pydash.utilities import to_path

    if key_path is not None:
        key_path_tokens = to_path(key_path)
        parent_node = _RefItem(obj, None)
        for tk in key_path_tokens:
            if isinstance(tk, int):
                if parent_node.is_none():
                    parent_node.set_(default_sequence_factory())
                if len(parent_node) <= tk:
                    parent_node.extend([None] * (tk + 1 - len(parent_node)))
                parent_node = _RefItem(parent_node.get_(), tk)
            else:
                if parent_node.is_none():
                    parent_node.set_(default_mapping_factory())
                if tk not in parent_node:
                    parent_node[tk] = None
                parent_node = _RefItem(parent_node.get_(), tk)

        if append and not parent_node.is_none():
            curr_val = parent_node.get_()
            if isinstance(curr_val, (bytes, str)) or not isinstance(
                curr_val, t.Sequence
            ):
                curr_val = [curr_val]
            if isinstance(value, (bytes, str)) or not isinstance(value, t.Sequence):
                value = [value]
            value = default_sequence_factory(list(curr_val) + list(value))
        parent_node.set_(value)



import itertools
from pydantic import BaseModel
from typing import Optional, Sequence, Mapping, Set, Iterable, Collection, Any

import networkx as nx

from pipelime.piper.model import DAGModel, PipelimeCommand


class GraphNode:
    GRAPH_NODE_TYPE_OPERATION = "operation"
    GRAPH_NODE_TYPE_DATA = "data"

    def __init__(self, name: str, type: str):
        self.name = str(name)
        self.type = str(type)

    @property
    def id(self):
        return f"{self.type}({self.name})"

    def __hash__(self) -> int:
        return hash(f"{self.type}({self.name})")

    def __repr__(self) -> str:
        return f"{self.type}({self.name})"

    def __eq__(self, o) -> bool:
        return self.id == o.id


class GraphNodeOperation(GraphNode):
    def __init__(self, name: str, command: PipelimeCommand):
        super().__init__(name, GraphNode.GRAPH_NODE_TYPE_OPERATION)
        self._command = command

    @property
    def command(self) -> PipelimeCommand:
        return self._command


class GraphNodeData(GraphNode):
    def __init__(self, name: str, path: str):
        super().__init__(name, GraphNode.GRAPH_NODE_TYPE_DATA)
        self._path = str(path)

    @property
    def path(self) -> str:
        return self._path


class DAGNodesGraph:
    class GraphAttrs:
        INPUT_PORT = "input_port"
        OUTPUT_PORT = "output_port"
        EDGE_TYPE = "edge_type"

    def __init__(self, raw_graph: nx.DiGraph):
        """Initialize the DAG graph starting from a raw directed graph (networkx).

        Args:
            raw_graph (nx.DiGraph): The raw graph to be converted to a DAG graph.
        """
        super().__init__()
        self._raw_graph = raw_graph

    @property
    def raw_graph(self) -> nx.DiGraph:
        """The raw graph (networkx) of the DAG.

        Returns:
            nx.DiGraph: The raw graph (networkx) of the DAG.
        """
        return self._raw_graph

    @property
    def operations_graph(self) -> "DAGNodesGraph":
        """The operations graph (networkx) of the DAG. It is a filtered version of the
        raw graph with operations nodes only.

        Returns:
            DAGNodesGraph: The operations graph (networkx) of the DAG.
        """
        return DAGNodesGraph.filter_node_graph(
            self, [GraphNode.GRAPH_NODE_TYPE_OPERATION]
        )

    @property
    def data_graph(self) -> "DAGNodesGraph":
        """The data graph (networkx) of the DAG. It is a filtered version of the
        raw graph with data nodes only.

        Returns:
            DAGNodesGraph: The data graph (networkx) of the DAG.
        """
        return DAGNodesGraph.filter_node_graph(self, [GraphNode.GRAPH_NODE_TYPE_DATA])

    @property
    def root_nodes(self) -> Sequence[GraphNode]:
        """The root nodes of the DAG. They are the nodes that have no predecessors.

        Returns:
            Sequence[GraphNode]: The root nodes of the DAG.
        """
        return [
            node
            for node in self.raw_graph.nodes
            if len(list(self.raw_graph.predecessors(node))) == 0
        ]

    def consumable_operations(
        self,
        produced_data: Set[GraphNodeData],
    ) -> Set[GraphNodeOperation]:
        """Given a set of produced data, returns the set of operations that can be
        consumed given the produced data, i.e. the operations that have inputs available

        Args:
            produced_data (Set[GraphNodeData]): The set of produced data.

        Returns:
            Set[GraphNodeOperation]: The set of operations that can be consumed given
            the produced data.
        """

        consumables = set()
        for node in self.operations_graph.raw_graph.nodes:
            in_data = [
                x
                for x in self.raw_graph.predecessors(node)
                if isinstance(x, GraphNodeData)
            ]
            if all(x in produced_data for x in in_data):
                consumables.add(node)
        return consumables

    def consume(
        self,
        operation_nodes: Collection[GraphNodeOperation],
    ) -> Set[GraphNodeData]:
        """Given a set of operations, consume them and return the set of produced data,
        i.e. the data are the outputs of the operations.

        Args:
            operation_nodes (Collection[GraphNodeOperation]): The set of operations
                to be consumed.

        Returns:
            Set[GraphNodeData]: The set of produced data.
        """
        consumed_data = set()
        for node in operation_nodes:
            out_data = [
                x
                for x in self.raw_graph.successors(node)
                if isinstance(x, GraphNodeData)
            ]
            consumed_data.update(out_data)
        return consumed_data

    def build_execution_stack(self) -> Sequence[Set[GraphNodeOperation]]:
        """Builds the execution stack of the DAG. The execution stack is a list of lists
        of operations. Each list of operations is a virtual layer of operations that
        can be executed in parallel because they have no dependencies.

        Returns:
            Sequence[Sequence[GraphNodeOperation]]: The execution stack of the DAG.
        """

        # the execution stack is a list of lists of nodes. Each list represents a
        execution_stack: Sequence[Set[GraphNodeOperation]] = []

        # initalize global produced data with the root nodes
        global_produced_data = set()
        global_produced_data.update(
            [x for x in self.root_nodes if isinstance(x, GraphNodeData)]
        )

        # set of operations that have been consumed
        consumed_operations = set()

        while True:
            # which operations can be consumed given the produced data?
            consumable: set = self.consumable_operations(global_produced_data)

            # Remove from consumable operations the ones that have already been consumed
            consumable = consumable.difference(consumed_operations)

            # No consumable operations? We are done!
            if len(consumable) == 0:
                break

            # If not empty, append consumable operations to the execution stack
            execution_stack.append(consumable)

            # The set of produced data is the union of all the consumed operations
            produced_data: set = self.consume(consumable)

            # Add the consumed operations to the consumed operations set
            consumed_operations.update(consumable)

            # Add the produced data to the global produced data
            global_produced_data.update(produced_data)

        return execution_stack

    @classmethod
    def build_nodes_graph(
        cls,
        dag_model: DAGModel,
    ) -> "DAGNodesGraph":
        """Builds the nodes graph of the DAG starting from a plain DAG model.

        Returns:
            DAGNodesGraph: The nodes graph of the DAG.
        """

        g = nx.DiGraph()

        for node_name, node in dag_model.nodes.items():
            node: PipelimeCommand

            inputs = node.get_inputs()
            outputs = node.get_outputs()

            cls._add_io_ports(g, node_name, node, inputs, True)
            cls._add_io_ports(g, node_name, node, outputs, False)

        return DAGNodesGraph(raw_graph=g)

    @classmethod
    def _add_io_ports(
        cls,
        di_graph: nx.DiGraph,
        node_name: str,
        node_cmd: PipelimeCommand,
        io_map: Optional[Mapping[str, Any]],
        is_input: bool,
    ):
        def _to_str(x):
            if isinstance(x, (str, bytes)):
                return str(x)
            if hasattr(x, "__piper_repr__"):
                return x.__piper_repr__()
            return repr(x)

        if io_map is not None:
            for name, value in io_map.items():
                if (
                    isinstance(value, str)
                    or isinstance(value, BaseModel)
                    or not isinstance(value, Iterable)
                ):
                    value = [value]

                attrs = {}
                for x in value:
                    if x:  # discard empty strings and None
                        x_name = _to_str(x)
                        n0 = GraphNodeData(x_name, x_name)
                        n1 = GraphNodeOperation(node_name, node_cmd)
                        if is_input:
                            port_attr = DAGNodesGraph.GraphAttrs.INPUT_PORT
                            edge_attr = "DATA_2_OPERATION"
                        else:
                            port_attr = DAGNodesGraph.GraphAttrs.OUTPUT_PORT
                            edge_attr = "OPERATION_2_DATA"
                            n0, n1 = n1, n0

                        di_graph.add_edge(n0, n1)
                        attrs.update(
                            {
                                (n0, n1): {
                                    port_attr: name,
                                    DAGNodesGraph.GraphAttrs.EDGE_TYPE: edge_attr,
                                }
                            }
                        )
                nx.set_edge_attributes(di_graph, attrs)

    @classmethod
    def filter_raw_graph(
        cls,
        full_graph: nx.DiGraph,
        types: Sequence[str],
    ) -> nx.DiGraph:
        """Filters the raw graph of the DAG to keep only the nodes of the given types.

        Args:
            full_graph (nx.DiGraph): The raw graph of the DAG.
            types (Sequence[str]): The types of the nodes to keep.

        Returns:
            nx.DiGraph: The filtered raw graph of the DAG.
        """

        filtered_graph: nx.DiGraph = nx.DiGraph()

        for node in full_graph.nodes:
            if node.type not in types:
                predecessors = list(full_graph.predecessors(node))
                successors = list(full_graph.successors(node))
                pairs = list(itertools.product(predecessors, successors))
                for pair in pairs:
                    filtered_graph.add_edge(pair[0], pair[1])

        # There could be single layers graph without connections,
        # all nodes have to be kept and are therefore added to the graph
        for node in full_graph.nodes:
            if node.type in types:
                filtered_graph.add_node(node)

        return filtered_graph

    @classmethod
    def filter_node_graph(
        cls,
        nodes_graph: "DAGNodesGraph",
        types: Sequence[str],
    ) -> "DAGNodesGraph":
        """Filters the nodes graph of the DAG to keep only the nodes of the given types.
        It wraps the raw graph into a DAGNodesGraph.

        Args:
            nodes_graph (DAGNodesGraph): The nodes graph of the DAG.
            types (Sequence[str]): The types of the nodes to keep.

        Returns:
            DAGNodesGraph: The filtered nodes graph of the DAG.
        """
        return DAGNodesGraph(
            raw_graph=cls.filter_raw_graph(nodes_graph.raw_graph, types)
        )



from pipelime.piper.progress.tracker.factory import TrackCallbackFactory
from pipelime.piper.progress.tracker.base import TrackCallback


class TestTrackCallbackFactory:
    def test_get_callback(self):
        callback = TrackCallbackFactory.get_callback()
        assert isinstance(callback, TrackCallback)



from __future__ import annotations

import os
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Set

import pydash as py_

import pipelime.choixe.ast.nodes as ast
from pipelime.choixe.ast.parser import parse
from pipelime.choixe.utils.io import load


@dataclass
class Inspection:
    imports: Set[Path] = field(default_factory=set)
    variables: Dict[str, Any] = field(default_factory=dict)
    help_strings: Dict[str, Any] = field(default_factory=dict)
    environ: Dict[str, Any] = field(default_factory=dict)
    symbols: Set[str] = field(default_factory=set)
    processed: bool = False

    def _iteratee(self, obj_value, src_value, key, obj, source) -> Any:
        res = None
        if obj_value is not None and src_value is None:
            res = obj_value
        return res

    def __add__(self, other: Inspection) -> Inspection:
        return Inspection(
            imports={*self.imports, *other.imports},
            variables=py_.merge_with(
                self.variables, other.variables, iteratee=self._iteratee
            ),  # type: ignore
            environ={**self.environ, **other.environ},
            symbols={*self.symbols, *other.symbols},
            processed=self.processed and other.processed,
            help_strings=py_.merge_with(
                self.help_strings, other.help_strings, iteratee=self._iteratee
            ),  # type: ignore
        )


class Inspector(ast.NodeVisitor):
    def __init__(self, cwd: Optional[Path] = None) -> None:
        super().__init__()
        self._cwd = cwd if cwd is not None else Path(os.getcwd())
        self._named_for_loops: Dict[str, str] = {}

    def _ignore(self, node: ast.Node) -> Any:
        return Inspection(processed=True)

    def visit_dict(self, node: ast.DictNode) -> Inspection:
        inspections = []
        for k, v in node.nodes.items():
            inspections.append(k.accept(self))
            inspections.append(v.accept(self))
        return sum(inspections, start=Inspection(processed=True))

    def visit_list(self, node: ast.ListNode) -> Inspection:
        start = Inspection(processed=True)
        return sum([x.accept(self) for x in node.nodes], start=start)

    def visit_literal(self, node: ast.LiteralNode) -> Inspection:
        return Inspection(processed=True)

    def visit_dict_bundle(self, node: ast.DictBundleNode) -> Any:
        start = Inspection(processed=True)
        return sum([x.accept(self) for x in node.nodes], start=start)

    def visit_str_bundle(self, node: ast.StrBundleNode) -> Inspection:
        start = Inspection(processed=True)
        return sum([x.accept(self) for x in node.nodes], start=start)

    def visit_var(self, node: ast.VarNode) -> Inspection:
        id_insp = node.identifier.accept(self)
        default_insp = (
            node.default.accept(self) if node.default else Inspection(processed=True)
        )
        env_insp = node.env.accept(self) if node.env else Inspection(processed=True)
        insp = id_insp + default_insp + env_insp

        if insp.processed:
            id_ = node.identifier.data  # type: ignore
            default = None if node.default is None else node.default.data  # type: ignore
            variables = py_.set_({}, id_, default)

            help_strings = {}
            if node.help is not None:
                help_strings = py_.set_({}, id_, node.help.data)

            environ = {}
            if node.env is not None and node.env.data:  # type: ignore
                environ[id_] = default

            insp = insp + Inspection(
                variables=variables, environ=environ, help_strings=help_strings
            )

        return insp

    def visit_import(self, node: ast.ImportNode) -> Inspection:
        insp = node.path.accept(self)

        if insp.processed:
            path = Path(node.path.data)  # type: ignore
            if not path.is_absolute():
                path = self._cwd / path

            if path.exists():
                subdata = load(path)
                parsed = parse(subdata)

                old_cwd = self._cwd
                self._cwd = path.parent
                nested = parsed.accept(self)
                self._cwd = old_cwd
            else:
                warnings.warn(f"Cannot complete inspection: file {path} is missing.")
                nested = Inspection()

            insp = Inspection(imports={Path(path).resolve()}) + nested + insp

        return insp

    def visit_sweep(self, node: ast.SweepNode) -> Inspection:
        start = Inspection()
        return sum([x.accept(self) for x in node.cases], start=start)

    def visit_symbol(self, node: ast.SymbolNode) -> Any:
        insp = node.symbol.accept(self)
        if insp.processed:
            insp = insp + Inspection(symbols={str(node.symbol.data)})  # type: ignore
        return insp

    def visit_instance(self, node: ast.InstanceNode) -> Inspection:
        insp = node.symbol.accept(self)
        if insp.processed:
            insp = insp + Inspection(symbols={str(node.symbol.data)})  # type: ignore
        return insp + node.args.accept(self)

    def visit_model(self, node: ast.ModelNode) -> Inspection:
        return self.visit_instance(node)

    def visit_for(self, node: ast.ForNode) -> Inspection:
        if node.identifier is not None:
            self._named_for_loops[node.identifier.data] = node.iterable.data
        iterable_insp = Inspection(variables=py_.set_({}, node.iterable.data, None))
        body_insp = node.body.accept(self)
        return iterable_insp + body_insp

    def visit_switch(self, node: ast.SwitchNode) -> Inspection:
        insp = node.value.accept(self)
        if insp.processed:
            insp = Inspection(variables=py_.set_({}, node.value.data, None))  # type: ignore
        cases_insp = sum(
            [x.accept(self) + y.accept(self) for x, y in node.cases],
            start=Inspection(processed=True),
        )
        default_insp = (
            node.default.accept(self) if node.default else Inspection(processed=True)
        )
        return insp + cases_insp + default_insp

    def visit_index(self, node: ast.IndexNode) -> Inspection:
        insp = Inspection()
        if node.identifier is not None:
            insp = insp + node.identifier.accept(self)
        return insp

    def visit_item(self, node: ast.ItemNode) -> Inspection:
        insp = Inspection()
        variables = {}
        if node.identifier is not None:
            sub_insp = node.identifier.accept(self)
            if sub_insp.processed:
                loop_id, _, key = node.identifier.data.partition(".")  # type: ignore
                iterable_name = self._named_for_loops[loop_id]
                full_path = f"{iterable_name}.{key}"
                py_.set_(variables, full_path, None)
            insp = insp + sub_insp + Inspection(variables=variables)
        return insp


def inspect(node: ast.Node, cwd: Optional[Path] = None) -> Inspection:
    inspector = Inspector(cwd=cwd)
    return node.accept(inspector)



from pipelime.sequences import SamplesSequence, Sample
import pydantic as pyd


class PipedSequenceBase(SamplesSequence):
    """Reasonable base implementation of `size` and `get_sample`."""

    # subclasses may override and give a proper description
    source: SamplesSequence = pyd.Field(
        ..., description="The source sample sequence.", exclude=True, pipe_source=True
    )

    def size(self) -> int:
        return len(self.source)

    def get_sample(self, idx: int) -> Sample:
        return self.source[idx]



from typing import Sequence

from pipelime.piper.drawing.base import NodesGraphDrawer
from pipelime.piper.drawing.graphviz import GraphvizNodesGraphDrawer
from pipelime.piper.drawing.mermaid import MermaidNodesGraphDrawer


class NodesGraphDrawerFactory:

    FACTORY_MAP = {
        "graphviz": GraphvizNodesGraphDrawer,
        "mermaid": MermaidNodesGraphDrawer,
    }

    @classmethod
    def create(cls, backend: str) -> NodesGraphDrawer:
        if backend not in cls.FACTORY_MAP:
            raise ValueError(f"Backend {backend} not supported")
        return cls.FACTORY_MAP[backend]()

    @classmethod
    def available_backends(cls) -> Sequence[str]:
        return list(cls.FACTORY_MAP.keys())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union

from pipelime.choixe.ast.nodes import Node
from pipelime.choixe.visitors.unparser import Unparser


@dataclass
class Entry:
    key: Sequence[Union[str, int]]
    value: Any

    def prepend(self, key) -> Entry:
        return Entry([key, *self.key], self.value)


@dataclass
class Chunk:
    entries: Sequence[Entry]

    def prepend(self, key: Union[int, str]) -> Chunk:
        return Chunk([x.prepend(key) for x in self.entries])

    def __add__(self, other: Chunk) -> Chunk:
        return Chunk(list(self.entries) + list(other.entries))


# 🏜️🤠🌵
class Walker(Unparser):
    """Specialization of the `Unparser` for the walk operation."""

    def __init__(self) -> None:
        super().__init__()

        # Wrap all visiting methods adding a chunkify step.
        for x in dir(self):
            if x.startswith("visit_"):
                setattr(self, x, self._wrapped_visit(getattr(self, x)))

    def _chunkify(self, data: Dict) -> Chunk:
        chunk = Chunk([])

        for k, v in data.items():
            if not isinstance(v, Chunk):
                v = Chunk([Entry([], v)])

            chunk += v.prepend(k)

        return chunk

    def _wrapped_visit(self, func: Callable[[Node], Any]) -> Callable[[Node], Any]:
        def wrapper(node: Node) -> Chunk:
            data = func(node)

            if isinstance(data, List):
                data = dict(enumerate(data))

            if isinstance(data, Dict):
                return self._chunkify(data)
            else:
                return data

        return wrapper


def walk(node: Node) -> List[Tuple[List[Union[str, int]], Any]]:
    """A special unparsing operation that also flattens a structure into a list of
    (deep_key, value) tuples. A deep key is a list of either strings or integers, that
    can be used as a pydash path.

    Example::

        data = {
            "a": 10,
            "b": [
                "bob",
                {
                    "a": 105,
                    "b": "alice",
                }
            ]
        }
        walk(parse(data))
        # [
        #     (["a"], 10),
        #     (["b", 0], "bob"),
        #     (["b", 1, "a"], 105),
        #     (["b", 1, "b"], "alice"),
        # ]

    Args:
        node (Node): The Choixe AST node to walk.

    Returns:
        List[Tuple[List[Union[str, int]], Any]]: The flattened unparsed node.
    """
    chunk: Chunk = node.accept(Walker())
    if not isinstance(chunk, Chunk):
        return [([], chunk)]
    return [(list(x.key), x.value) for x in chunk.entries]



from loguru import logger

from pipelime.piper.progress.model import ProgressUpdate
from pipelime.piper.progress.tracker.base import TrackCallback


class LoguruTrackCallback(TrackCallback):
    """Loguru tracker callback"""

    def __init__(self, level: str = "INFO") -> None:
        super().__init__()
        self._level = level

    def on_start(self, prog: ProgressUpdate) -> None:
        logger.log(
            self._level,
            "Token: {} | Node: {} | Chunk: {} | {} | Started.",
            prog.op_info.token,
            prog.op_info.node,
            prog.op_info.chunk,
            prog.op_info.message,
        )

    def on_advance(self, prog: ProgressUpdate):
        logger.log(
            self._level,
            "Token: {} | Node: {} | Chunk: {} | {} | Advanced of {} steps.",
            prog.op_info.token,
            prog.op_info.node,
            prog.op_info.chunk,
            prog.op_info.message,
            prog.progress,
        )

    def on_finish(self, prog: ProgressUpdate) -> None:
        logger.log(
            self._level,
            "Token: {} | Node: {} | Chunk: {} | {} | Finished.",
            prog.op_info.token,
            prog.op_info.node,
            prog.op_info.chunk,
            prog.op_info.message,
        )



from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def piper_commands(data_folder):

    base_path = Path(data_folder) / "piper" / "piper_commands"

    def build_python_command(command: str):
        return f"python {str(base_path / command)}"

    return {
        "fake_detector": {
            "command": build_python_command("fake_detector.py"),
            "valid": True,
            "exception": None,
        },
    }


@pytest.fixture(scope="session")
def piper_dags(data_folder, piper_commands, tmp_path_factory: pytest.TempPathFactory):

    base_path = Path(data_folder) / "piper" / "dags"

    return {
        "complex": {
            "folder": base_path / "complex",
            "valid": True,
            "exception": None,
            "graph": True,
            "executable": {
                # If success is TRUE a "final_validation.py" file is needed in the
                # dag folder
                "success": True,
                "executable_placeholders": {
                    "CUSTOM_COMMAND": piper_commands["fake_detector"]["command"],
                    "OUTPUT_FOLDER": tmp_path_factory.mktemp("output_folder"),
                    "FINAL_FOLDER": tmp_path_factory.mktemp("final_folder"),
                    "SIZE": 10,
                    "SCHEMA_GENERATED": base_path
                    / "complex"
                    / "schema_generated.schema",
                    "SCHEMA_REMAPPED": base_path / "complex" / "schema_remapped.schema",
                    "SCHEMA_DETECTED": base_path / "complex" / "schema_detected.schema",
                },
                "exception": None,
            },
        },
        # "validation_error": {
        #     "folder": base_path / "validation_error",
        #     "valid": True,
        #     "exception": None,
        #     "graph": True,
        #     "executable": {
        #         # If success is TRUE a "final_validation.py" file is needed in the
        #         # dag folder
        #         "success": False,
        #         "executable_placeholders": {
        #             "OUTPUT_FOLDER": tmp_path_factory.mktemp("output_folder"),
        #             "SIZE": 10,
        #             "SCHEMA_GENERATED": base_path
        #             / "validation_error"
        #             / "schema_generated.schema",
        #         },
        #         "exception": SampleSchema.ValidationError,
        #     },
        # },
        # "validation_error_noschema_file": {
        #     "folder": base_path / "validation_error_noschema_file",
        #     "valid": True,
        #     "exception": None,
        #     "graph": True,
        #     "executable": {
        #         # If success is TRUE a "final_validation.py" file is needed in the
        #         # dag folder
        #         "success": False,
        #         "executable_placeholders": {
        #             "OUTPUT_FOLDER": tmp_path_factory.mktemp("output_folder"),
        #             "SIZE": 10,
        #             "SCHEMA_GENERATED": base_path
        #             / "validation_error_noschema_file"
        #             / "schema_generated.schema",
        #         },
        #         "exception": SampleSchema.ValidationError,
        #     },
        # },
        # "no_params": {
        #     "folder": base_path / "no_params",
        #     "valid": True,
        #     "exception": None,
        #     "graph": False,
        # },
        # "invalid_params": {
        #     "folder": base_path / "invalid_params",
        #     "valid": False,
        #     "exception": KeyError,
        #     "graph": False,
        # },
        # "invalid_nodes": {
        #     "folder": base_path / "invalid_nodes",
        #     "valid": False,
        #     "exception": pydantic.ValidationError,
        #     "graph": False,
        # },
        # "invalid_node_foreach": {
        #     "folder": base_path / "invalid_node_foreach",
        #     "valid": False,
        #     "exception": KeyError,
        #     "graph": False,
        # },
        # "invalid_node_foreach_do_content": {
        #     "folder": base_path / "invalid_node_foreach_do_content",
        #     "valid": False,
        #     "exception": TypeError,
        #     "graph": False,
        # },
        # "invalid_arg_foreach": {
        #     "folder": base_path / "invalid_arg_foreach",
        #     "valid": False,
        #     "exception": KeyError,
        #     "graph": False,
        # },
        # "invalid_arg_foreach_content": {
        #     "folder": base_path / "invalid_arg_foreach_content",
        #     "valid": False,
        #     "exception": TypeError,
        #     "graph": False,
        # },
        # "single_operation_with_multi_branches": {
        #     "folder": base_path / "single_operation_with_multi_branches",
        #     "valid": True,
        #     "exception": None,
        #     "graph": True,
        #     "executable": {
        #         # If success is TRUE a "final_validation.py" file is needed in the
        #         # dag folder
        #         "success": True,
        #         "executable_placeholders": {
        #             "OUTPUT_FOLDER": tmp_path_factory.mktemp("output_folder"),
        #         },
        #         "exception": None,
        #     },
        # },
    }



from __future__ import annotations
from pipelime.sequences import SamplesSequence
import typing as t


class DirectAccessSequence(t.Sequence[t.Mapping[str, t.Any]]):
    def __init__(self, sequence: SamplesSequence):
        self._sequence = sequence

    def __len__(self) -> int:
        return len(self._sequence)

    def __getitem__(self, idx: t.Union[int, slice]) -> t.Mapping[str, t.Any]:
        return (
            DirectAccessSequence(self._sequence[idx])  # type: ignore
            if isinstance(idx, slice)
            else self._sequence[idx].to_dict()
        )

    def __add__(
        self, other: t.Union[DirectAccessSequence, SamplesSequence]
    ) -> DirectAccessSequence:
        if isinstance(other, DirectAccessSequence):
            return DirectAccessSequence(
                self._sequence.cat(other._sequence)  # type: ignore
            )
        return DirectAccessSequence(self._sequence.cat(other))  # type: ignore



import pytest
from pathlib import Path
import numpy as np
import typing as t

import pipelime.items as pli


def _np_eq(x, y) -> bool:
    return np.array_equal(x, y, equal_nan=True)


def _np_eq_1d(x, y) -> bool:
    return np.array_equal(np.atleast_1d(x), y, equal_nan=True)


class TestItems:
    def test_reference_items(self, items_folder: Path):
        from pipelime.items.base import ItemFactory

        checked_items = set()
        for fp in items_folder.iterdir():
            if fp not in checked_items:
                checked_items.add(fp)
                ref_item = ItemFactory.get_instance(fp)()
                for other in items_folder.glob(fp.stem + ".*"):
                    if other not in checked_items:
                        checked_items.add(other)
                        trg_item = ItemFactory.get_instance(other)()

                        if isinstance(ref_item, np.ndarray):
                            assert _np_eq(ref_item, trg_item)
                        elif isinstance(ref_item, tuple):  # the pickle file is a tuple
                            assert ref_item[3] == trg_item
                        elif isinstance(trg_item, tuple):
                            assert ref_item == trg_item[3]
                        else:
                            assert ref_item == trg_item

    @pytest.mark.parametrize(
        ["item_cls", "value", "eq_fn"],
        [
            (pli.PickleItem, (42, "asdf", 3.14), lambda x, y: x == y),
            (pli.BinaryItem, b"asdfiasodifoj123124214", lambda x, y: x == y),
            (pli.NpyNumpyItem, (np.random.rand(3, 4) * 100), _np_eq),
            (pli.TxtNumpyItem, (np.random.rand(3, 4) * 100), _np_eq_1d),
            (pli.NpyNumpyItem, 42.42, _np_eq),
            (pli.TxtNumpyItem, 42.42, _np_eq_1d),
            (pli.BmpImageItem, (np.random.rand(3, 4) * 100).astype(np.uint8), _np_eq),
            (pli.PngImageItem, (np.random.rand(3, 4) * 100).astype(np.uint8), _np_eq),
            (pli.TiffImageItem, (np.random.rand(3, 4) * 100).astype(np.uint8), _np_eq),
            (
                pli.JsonMetadataItem,
                {"a": [1, 2, 3], "b": 3.14, "c": [True, False]},
                lambda x, y: x == y,
            ),
            (
                pli.YamlMetadataItem,
                {"a": [1, 2, 3], "b": 3.14, "c": [True, False]},
                lambda x, y: x == y,
            ),
            (
                pli.TomlMetadataItem,
                {"a": [1, 2, 3], "b": 3.14, "c": [True, False]},
                lambda x, y: x == y,
            ),
        ],
    )
    def test_read_write(self, tmp_path: Path, item_cls: t.Type[pli.Item], value, eq_fn):
        value_path = tmp_path / "value"

        w_item = item_cls(value)
        w_item.serialize(value_path)
        assert eq_fn(value, w_item())

        r_item = item_cls(value_path.with_suffix(item_cls.file_extensions()[0]))
        assert eq_fn(w_item(), r_item())

    def test_read_write_jpeg(self, tmp_path: Path):
        import imageio.v3 as iio
        from io import BytesIO

        data_path = tmp_path / "data"
        data = (np.random.rand(3, 4) * 100).astype(np.uint8)

        w_item = pli.JpegImageItem(data)
        w_item.serialize(data_path)
        assert _np_eq(data, w_item())

        jpeg_suffix = pli.JpegImageItem.file_extensions()[0]
        r_item = pli.JpegImageItem(data_path.with_suffix(jpeg_suffix))

        encoded = BytesIO()
        iio.imwrite(
            encoded, data, extension=jpeg_suffix, **pli.JpegImageItem.save_options()
        )
        encoded.seek(0)
        decoded = np.array(iio.imread(encoded, extension=jpeg_suffix))

        assert _np_eq(r_item(), decoded)

    def test_data_cache(self, items_folder: Path):
        from pipelime.items.base import ItemFactory

        for v in ItemFactory.ITEM_DATA_CACHE_MODE.values():
            assert v is None

        def _reload() -> t.Tuple[pli.Item, pli.Item]:
            bmp_item = ItemFactory.get_instance(items_folder / "0.bmp")
            json_item = ItemFactory.get_instance(items_folder / "3.json")
            return bmp_item, json_item

        def _check_single(item, should_cache, is_cached):
            item.cache_data = should_cache
            assert item.cache_data is should_cache
            _ = item()
            if is_cached:
                assert item._data_cache is not None
            else:
                assert item._data_cache is None

        def _check(should_cache, is_cached):
            for id, it in enumerate(_reload()):
                _check_single(it, should_cache[id], is_cached[id])

        _check((True, True), (True, True))
        _check((False, False), (False, False))

        with pli.no_data_cache(pli.NumpyItem):
            _check((True, True), (True, True))
            _check((None, None), (False, True))

        with pli.no_data_cache(pli.NumpyItem, pli.JsonMetadataItem):
            pli.enable_item_data_cache(pli.NumpyItem)
            _check((True, True), (True, True))
            _check((False, True), (False, True))
            _check((None, None), (True, False))
            _check((None, None), (True, False))

    def test_set_data_twice(self):
        with pytest.raises(ValueError):
            _ = pli.NpyNumpyItem(np.random.rand(3, 4), np.random.rand(3, 4))

    def test_set_from_binary(self, items_folder: Path):
        with (items_folder / "2.txt").open("rb") as fp:
            ref = pli.TxtNumpyItem(fp)()
            gt = np.array(
                [
                    [
                        6.390370900299208179e-01,
                        4.262250008515110489e-01,
                        9.315973628491435177e-01,
                        1.002942906369641562e-01,
                    ],
                    [
                        7.517292051086260640e-01,
                        1.087911338619208523e-01,
                        9.393092045579851668e-01,
                        2.912748772468053415e-01,
                    ],
                    [
                        3.411277654275945981e-01,
                        8.135318888374636348e-01,
                        7.341307526247440318e-01,
                        1.800992012876649895e-01,
                    ],
                ]
            )

            assert isinstance(ref, np.ndarray)
            assert gt.shape == ref.shape
            assert np.all(gt - ref < 1e-6)  # type: ignore

    def test_invalid_ext(self):
        with pytest.raises(ValueError):
            _ = pli.NpyNumpyItem(Path("foo.bar"))

    def test_invalid_source(self):
        from urllib.parse import ParseResult

        item = pli.NpyNumpyItem(
            Path("foo.npy"),
            ParseResult(
                scheme="foo",
                netloc="bar",
                path="baz.npy",
                params="",
                query="",
                fragment="",
            ),
        )
        assert item() is None

    def test_disabled_serialization_modes(  # noqa: C901
        self, minimnist_private_dataset: dict, tmp_path: Path
    ):
        import pipelime.items as pli
        from pipelime.sequences import SamplesSequence
        from pipelime.stages import StageUploadToRemote
        from pipelime.remotes import make_remote_url

        # data lake
        remote_url = make_remote_url(
            scheme="file",
            host="localhost",
            bucket=(tmp_path / "rmbucket"),
        )

        input_seq = SamplesSequence.from_underfolder(  # type: ignore
            minimnist_private_dataset["path"], merge_root_items=True
        )

        # upload to remote (remote source is added to the items)
        for _ in input_seq.map(
            StageUploadToRemote(remotes=[remote_url])  # type: ignore
        ):
            pass

        # default mode: writing remote files
        for _ in input_seq.to_underfolder(folder=(tmp_path / "output_rm")):
            pass
        for p in (tmp_path / "output_rm").rglob("*"):
            if p.is_file():
                assert p.suffix == pli.Item.REMOTE_FILE_EXT

        # disable writing remote files
        with pli.item_disabled_serialization_modes("REMOTE_FILE"):
            for _ in input_seq.to_underfolder(folder=(tmp_path / "output_hl")):
                pass
            for p in (tmp_path / "output_hl").rglob("*"):
                if p.is_file():
                    assert p.suffix != pli.Item.REMOTE_FILE_EXT
                    assert not p.is_symlink()
                    assert p.stat().st_nlink > 1

        with pli.item_disabled_serialization_modes(
            pli.SerializationMode.REMOTE_FILE, pli.NumpyItem
        ):
            with pli.item_disabled_serialization_modes(
                [pli.SerializationMode.DEEP_COPY, "HARD_LINK"], pli.ImageItem
            ):
                for _ in input_seq.to_underfolder(folder=(tmp_path / "output_mx")):
                    pass
                for p in (tmp_path / "output_mx").rglob("*"):
                    if p.is_file():
                        assert not p.is_symlink()
                        if p.suffix == pli.TxtNumpyItem.file_extensions()[0]:
                            assert p.stat().st_nlink > 1
                        elif p.suffix == pli.PngImageItem.file_extensions()[0]:
                            assert p.stat().st_nlink == 1
                        else:
                            format_suffix = (
                                pli.JsonMetadataItem.file_extensions()[0]
                                if ".json" in p.suffixes
                                else pli.YamlMetadataItem.file_extensions()[0]
                            )
                            assert (
                                "".join(p.suffixes)
                                == format_suffix + pli.Item.REMOTE_FILE_EXT
                            )



import os
import typing as t
from pathlib import Path

import pydantic as pyd

import pipelime.sequences as pls


@pls.source_sequence
class UnderfolderReader(pls.SamplesSequence, title="from_underfolder"):
    """A SamplesSequence loading data from an Underfolder dataset."""

    folder: Path = pyd.Field(
        ..., description="The root folder of the Underfolder dataset."
    )
    merge_root_items: bool = pyd.Field(
        True,
        description=(
            "Adds root items as shared items "
            "to each sample (sample values take precedence)."
        ),
    )
    must_exist: bool = pyd.Field(
        True, description="If True raises an error when `folder` does not exist."
    )
    watch: bool = pyd.Field(
        False,
        description=(
            "If True, the dataset is scanned every time a new Sample is requested."
        ),
    )

    _samples: t.List[t.Union[pls.Sample, t.Dict[str, Path]]] = pyd.PrivateAttr(
        default_factory=list
    )
    _root_sample: t.Optional[t.Union[pls.Sample, t.Dict[str, Path]]] = pyd.PrivateAttr(
        None
    )

    @pyd.validator("must_exist", always=True)
    def check_folder_exists(cls, v, values, **kwargs):
        p = values["folder"]
        if v and not p.exists():
            raise ValueError(f"Root folder {p} does not exist.")
        return v

    def __init__(self, folder: Path, **data):
        super().__init__(folder=folder, **data)  # type: ignore

        if not self.watch:
            self._scan_root_files()

        # load samples even if we are `watch`ing, so that we know the initial length
        self._scan_sample_files()

    @property
    def root_sample(self) -> pls.Sample:
        from pipelime.items.base import ItemFactory

        # user may change the value of `self.watch` at any time,
        # so both checks are necessary
        if self.watch or self._root_sample is None:
            self._scan_root_files()
        if not isinstance(self._root_sample, pls.Sample):
            self._root_sample = pls.Sample(
                {
                    k: ItemFactory.get_instance(v, shared_item=True)
                    for k, v in self._root_sample.items()  # type: ignore
                }
            )
        return self._root_sample

    def _extract_key(self, name: str) -> str:
        return name.partition(".")[0]

    def _extract_id_key(self, name: str) -> t.Optional[t.Tuple[int, str]]:
        id_key_split = name.partition("_")
        if not id_key_split[2]:
            return None  # pragma: no cover
        try:
            return (int(id_key_split[0]), self._extract_key(id_key_split[2]))
        except ValueError:  # pragma: no cover
            return None

    def _scan_root_files(self):
        if self.folder.exists():
            root_items: t.Dict[str, Path] = {}
            with os.scandir(str(self.folder)) as it:
                for entry in it:
                    if entry.is_file():
                        key = self._extract_key(entry.name)
                        if key:
                            root_items[key] = entry.path
            self._root_sample = root_items if root_items else pls.Sample()
        else:
            self._root_sample = pls.Sample()

    def _scan_sample_files(self):
        data_folder = self.folder / "data"
        if data_folder.exists():
            self._samples = []
            with os.scandir(str(data_folder)) as it:
                for entry in it:
                    if entry.is_file():
                        id_key = self._extract_id_key(entry.name)
                        if id_key:
                            self._samples.extend(
                                [{} for _ in range(id_key[0] - len(self._samples) + 1)]
                            )
                            self._samples[id_key[0]][id_key[1]] = (  # type: ignore
                                entry.path
                            )

    def size(self) -> int:
        return len(self._samples)

    def get_sample(self, idx: int) -> pls.Sample:
        from pipelime.items.base import ItemFactory

        if self.watch:
            self._scan_sample_files()

        sample = self._samples[idx]
        if not isinstance(sample, pls.Sample):
            sample = pls.Sample(
                {
                    k: ItemFactory.get_instance(v, shared_item=False)
                    for k, v in sample.items()
                }
            )
            self._samples[idx] = sample
        return self.root_sample.merge(sample) if self.merge_root_items else sample



if __name__ == "__main__":
    from pathlib import Path
    from pipelime.sequences import Grabber, SamplesSequence, grab_all

    seq = SamplesSequence.from_underfolder(
        Path(__file__).resolve().absolute().parents[2]
        / "tests/sample_data/datasets/underfolder_minimnist"
    )
    seq = seq.shuffle(seed=42)[2:6]
    writer = seq.to_underfolder("./writer_output", exists_ok=True).enumerate()

    # here comes the magic
    grabber = Grabber(num_workers=4)  # type: ignore
    grab_all(
        grabber,
        writer,
        sample_fn=(
            lambda x: print("sample #", int(x["~idx"]()), sep="")  # type: ignore
        ),
    )
